@@ -30,19 +30,33 @@
  * with types changed to match glib types, since those are defined for us.
  * However, upstream FAAD is distributed with a broken header file that defined
  * these wrongly (in a way which was broken on 64 bit systems).
+ *
  * Upstream CVS still has the bug, but has also renamed all the public symbols
- * for Better Corporate Branding (or whatever), so we're screwed there.
+ * for Better Corporate Branding (or whatever), so we need to take that
+ * (FAAD_IS_NEAAC) into account as well.
  *
  * We must call them using these definitions. Most distributions now have the
  * corrected header file (they distribute a patch along with the source), 
- * but not all, hence this Truly Evil Hack. This hack will need updating if
- * upstream ever releases something with the new API.
+ * but not all, hence this Truly Evil Hack.
+ *
+ * Note: The prototypes don't need to be defined conditionaly, as the cpp will
+ * do that for us.
  */
+#ifdef FAAD_IS_NEAAC
+#define NeAACDecInit NeAACDecInit_no_definition
+#define NeAACDecInit2 NeAACDecInit2_no_definition
+#else
 #define faacDecInit faacDecInit_no_definition
 #define faacDecInit2 faacDecInit2_no_definition
+#endif
 #include "gstfaad.h"
+#ifdef FAAD_IS_NEAAC
+#undef NeAACDecInit
+#undef NeAACDecInit2
+#else
 #undef faacDecInit
 #undef faacDecInit2
+#endif
 
 extern long faacDecInit (faacDecHandle, guint8 *, guint32, guint32 *, guint8 *);
 extern int8_t faacDecInit2 (faacDecHandle, guint8 *, guint32,
@@ -51,6 +65,8 @@ extern int8_t faacDecInit2 (faacDecHandle, guint8 *, guint32,
 GST_DEBUG_CATEGORY_STATIC (faad_debug);
 #define GST_CAT_DEFAULT faad_debug
 
+#define MAX_DECODE_ERRORS 5
+
 static const GstElementDetails faad_details =
 GST_ELEMENT_DETAILS ("AAC audio decoder",
     "Codec/Decoder/Audio",
@@ -111,6 +127,7 @@ static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src",
 static void gst_faad_base_init (GstFaadClass * klass);
 static void gst_faad_class_init (GstFaadClass * klass);
 static void gst_faad_init (GstFaad * faad);
+static void gst_faad_dispose (GObject * object);
 
 static gboolean gst_faad_setcaps (GstPad * pad, GstCaps * caps);
 static GstCaps *gst_faad_srcgetcaps (GstPad * pad);
@@ -170,10 +187,13 @@ gst_faad_base_init (GstFaadClass * klass)
 static void
 gst_faad_class_init (GstFaadClass * klass)
 {
+  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GstElementClass *gstelement_class = GST_ELEMENT_CLASS (klass);
 
   parent_class = g_type_class_peek_parent (klass);
 
+  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_faad_dispose);
+
   gstelement_class->change_state = GST_DEBUG_FUNCPTR (gst_faad_change_state);
 }
 
@@ -192,6 +212,8 @@ gst_faad_init (GstFaad * faad)
   faad->bytes_in = 0;
   faad->sum_dur_out = 0;
   faad->packetised = FALSE;
+  faad->error_count = 0;
+  faad->segment = gst_segment_new ();
 
   faad->sinkpad = gst_pad_new_from_static_template (&sink_template, "sink");
   gst_element_add_pad (GST_ELEMENT (faad), faad->sinkpad);
@@ -213,6 +235,20 @@ gst_faad_init (GstFaad * faad)
   gst_element_add_pad (GST_ELEMENT (faad), faad->srcpad);
 }
 
+static void
+gst_faad_dispose (GObject * object)
+{
+  GstFaad *faad = GST_FAAD (object);
+
+  if (faad->segment) {
+    gst_segment_free (faad->segment);
+    faad->segment = NULL;
+  }
+
+  G_OBJECT_CLASS (parent_class)->dispose (object);
+
+}
+
 static void
 gst_faad_send_tags (GstFaad * faad)
 {
@@ -269,18 +305,23 @@ gst_faad_setcaps (GstPad * pad, GstCaps * caps)
   if ((value = gst_structure_get_value (str, "codec_data"))) {
     guint32 samplerate;
     guint8 channels;
+    guint8 *cdata;
+    guint csize;
 
     /* We have codec data, means packetised stream */
     faad->packetised = TRUE;
-    buf = GST_BUFFER (gst_value_get_mini_object (value));
+    buf = gst_value_get_buffer (value);
+
+    cdata = GST_BUFFER_DATA (buf);
+    csize = GST_BUFFER_SIZE (buf);
+
+    if (csize < 2)
+      goto wrong_length;
 
     /* someone forgot that char can be unsigned when writing the API */
-    if ((gint8) faacDecInit2 (faad->handle,
-            GST_BUFFER_DATA (buf), GST_BUFFER_SIZE (buf), &samplerate,
-            &channels) < 0) {
-      GST_DEBUG ("faacDecInit2() failed");
-      return FALSE;
-    }
+    if ((gint8) faacDecInit2 (faad->handle, cdata, csize, &samplerate,
+            &channels) < 0)
+      goto init_failed;
 
     GST_DEBUG_OBJECT (faad, "channels=%u, rate=%u", channels, samplerate);
 
@@ -329,6 +370,18 @@ gst_faad_setcaps (GstPad * pad, GstCaps * caps)
     gst_faad_send_tags (faad);
 
   return TRUE;
+
+  /* ERRORS */
+wrong_length:
+  {
+    GST_DEBUG_OBJECT (faad, "codec_data less than 2 bytes long");
+    return FALSE;
+  }
+init_failed:
+  {
+    GST_DEBUG_OBJECT (faad, "faacDecInit2() failed");
+    return FALSE;
+  }
 }
 
 
@@ -775,6 +828,8 @@ gst_faad_sink_event (GstPad * pad, GstEvent * event)
         GST_DEBUG ("Got NEWSEGMENT event in GST_FORMAT_TIME, passing on (%"
             GST_TIME_FORMAT " - %" GST_TIME_FORMAT ")", GST_TIME_ARGS (start),
             GST_TIME_ARGS (end));
+        gst_segment_set_newsegment (faad->segment, is_update, rate, fmt, start,
+            end, base);
       } else if (fmt == GST_FORMAT_BYTES) {
         gint64 new_start = 0;
         gint64 new_end = -1;
@@ -797,6 +852,9 @@ gst_faad_sink_event (GstPad * pad, GstEvent * event)
         event = gst_event_new_new_segment (is_update, rate,
             GST_FORMAT_TIME, new_start, new_end, new_start);
 
+        gst_segment_set_newsegment (faad->segment, is_update, rate,
+            GST_FORMAT_TIME, new_start, new_end, new_start);
+
         GST_DEBUG ("Sending new NEWSEGMENT event, time %" GST_TIME_FORMAT
             " - %" GST_TIME_FORMAT, GST_TIME_ARGS (new_start),
             GST_TIME_ARGS (new_end));
@@ -949,10 +1007,10 @@ gst_faad_src_query (GstPad * pad, GstQuery * query)
 
 
 static gboolean
-gst_faad_update_caps (GstFaad * faad, faacDecFrameInfo * info,
-    GstCaps ** p_caps)
+gst_faad_update_caps (GstFaad * faad, faacDecFrameInfo * info)
 {
   GstAudioChannelPosition *pos;
+  gboolean ret;
   GstCaps *caps;
 
   /* store new negotiation information */
@@ -982,14 +1040,10 @@ gst_faad_update_caps (GstFaad * faad, faacDecFrameInfo * info,
 
   GST_DEBUG ("New output caps: %" GST_PTR_FORMAT, caps);
 
-  if (!gst_pad_set_caps (faad->srcpad, caps)) {
-    gst_caps_unref (caps);
-    return FALSE;
-  }
+  ret = gst_pad_set_caps (faad->srcpad, caps);
+  gst_caps_unref (caps);
 
-  *p_caps = caps;
-
-  return TRUE;
+  return ret;
 }
 
 /*
@@ -1000,7 +1054,6 @@ gst_faad_update_caps (GstFaad * faad, faacDecFrameInfo * info,
  * subsequent syncpoints (similar to mp3 typefinding in
  * gst/typefind/) for ADTS because 12 bits isn't very reliable.
  */
-
 static gboolean
 gst_faad_sync (GstBuffer * buf, guint * off)
 {
@@ -1010,6 +1063,10 @@ gst_faad_sync (GstBuffer * buf, guint * off)
 
   GST_DEBUG ("Finding syncpoint");
 
+  /* check for too small a buffer */
+  if (size < 3)
+    return FALSE;
+
   /* FIXME: for no-sync, we go over the same data for every new buffer.
    * We should save the information somewhere. */
   for (n = 0; n < size - 3; n++) {
@@ -1075,6 +1132,55 @@ looks_like_valid_header (guint8 * input_data, guint input_size)
   return TRUE;
 }
 
+/*
+  clips buffer to currently configured segment. Returns FALSE if the buffer 
+  has to be dropped.
+*/
+
+static gboolean
+clip_outgoing_buffer (GstFaad * faad, GstBuffer * buffer)
+{
+  gint64 start, stop, cstart, cstop, diff;
+  gboolean res = TRUE;
+
+  if (faad->segment->format != GST_FORMAT_TIME)
+    goto beach;
+
+  start = GST_BUFFER_TIMESTAMP (buffer);
+  stop = start + GST_BUFFER_DURATION (buffer);
+
+  if (gst_segment_clip (faad->segment, GST_FORMAT_TIME,
+          start, stop, &cstart, &cstop)) {
+    diff = cstart - start;
+    if (diff > 0) {
+      GST_BUFFER_TIMESTAMP (buffer) = cstart;
+      GST_BUFFER_DURATION (buffer) -= diff;
+
+      /* time->frames->bytes */
+      diff =
+          faad->bps * faad->channels * GST_CLOCK_TIME_TO_FRAMES (diff,
+          faad->samplerate);
+      GST_BUFFER_DATA (buffer) += diff;
+      GST_BUFFER_SIZE (buffer) -= diff;
+    }
+    diff = cstop - stop;
+    if (diff > 0) {
+      GST_BUFFER_DURATION (buffer) -= diff;
+      /* time->frames->bytes */
+      diff =
+          faad->bps * faad->channels * GST_CLOCK_TIME_TO_FRAMES (diff,
+          faad->samplerate);
+      /* update size */
+      GST_BUFFER_SIZE (buffer) -= diff;
+    }
+  } else {
+    GST_DEBUG_OBJECT (faad, "buffer is outside configured segment");
+    res = FALSE;
+  }
+
+beach:
+  return res;
+}
 
 static GstFlowReturn
 gst_faad_chain (GstPad * pad, GstBuffer * buffer)
@@ -1085,7 +1191,6 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
   guchar *input_data;
   GstFaad *faad;
   GstBuffer *outbuf;
-  GstCaps *caps = NULL;
   faacDecFrameInfo info;
   void *out;
   gboolean run_loop = TRUE;
@@ -1117,6 +1222,7 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
 
   input_data = GST_BUFFER_DATA (buffer);
   input_size = GST_BUFFER_SIZE (buffer);
+
   if (!faad->packetised) {
     if (!gst_faad_sync (buffer, &sync_off)) {
       goto next;
@@ -1182,10 +1288,15 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
     out = faacDecDecode (faad->handle, &info, input_data + skip_bytes,
         input_size - skip_bytes);
 
-    if (info.error) {
+    if (info.error > 0) {
       guint32 rate;
       guint8 ch;
 
+      GST_DEBUG_OBJECT (faad, "decoding error: %s",
+          faacDecGetErrorMessage (info.error));
+
+      goto out;
+
       if (!faad->packetised)
         goto decode_error;
 
@@ -1208,12 +1319,22 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
       out = faacDecDecode (faad->handle, &info, input_data + skip_bytes,
           input_size - skip_bytes);
 
-      if (info.error)
-        goto decode_error;
+      if (info.error) {
+        faad->error_count++;
+        if (faad->error_count >= MAX_DECODE_ERRORS)
+          goto decode_error;
+        GST_DEBUG_OBJECT (faad,
+            "Failed to decode buffer: %s, count = %d, trying to resync",
+            faacDecGetErrorMessage (info.error), faad->error_count);
+        continue;
+      }
+
+      faad->error_count = 0;    /* all fine, reset error counter */
     }
 
     if (info.bytesconsumed > input_size)
       info.bytesconsumed = input_size;
+
     input_size -= info.bytesconsumed;
     input_data += info.bytesconsumed;
 
@@ -1234,7 +1355,7 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
       }
 
       if (fmt_change) {
-        if (!gst_faad_update_caps (faad, &info, &caps)) {
+        if (!gst_faad_update_caps (faad, &info)) {
           GST_ELEMENT_ERROR (faad, CORE, NEGOTIATION, (NULL),
               ("Setting caps on source pad failed"));
           ret = GST_FLOW_ERROR;
@@ -1247,10 +1368,12 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
         guint bufsize = info.samples * faad->bps;
         guint num_samples = info.samples / faad->channels;
 
+        GST_DEBUG_OBJECT (faad, "decoded %d samples", info.samples);
+
         /* note: info.samples is total samples, not per channel */
         ret =
-            gst_pad_alloc_buffer_and_set_caps (faad->srcpad, 0, bufsize, caps,
-            &outbuf);
+            gst_pad_alloc_buffer_and_set_caps (faad->srcpad, 0, bufsize,
+            GST_PAD_CAPS (faad->srcpad), &outbuf);
         if (ret != GST_FLOW_OK)
           goto out;
 
@@ -1266,12 +1389,16 @@ gst_faad_chain (GstPad * pad, GstBuffer * buffer)
         faad->sum_dur_out += GST_BUFFER_DURATION (outbuf);
         GST_OBJECT_UNLOCK (faad);
 
-        GST_LOG_OBJECT (faad, "pushing buffer, off=%" G_GUINT64_FORMAT ", ts=%"
-            GST_TIME_FORMAT, GST_BUFFER_OFFSET (outbuf),
-            GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)));
-        if ((ret = gst_pad_push (faad->srcpad, outbuf)) != GST_FLOW_OK &&
-            ret != GST_FLOW_NOT_LINKED)
-          goto out;
+        if (clip_outgoing_buffer (faad, outbuf)) {
+          GST_LOG_OBJECT (faad,
+              "pushing buffer, off=%" G_GUINT64_FORMAT ", ts=%" GST_TIME_FORMAT,
+              GST_BUFFER_OFFSET (outbuf),
+              GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)));
+          ret = gst_pad_push (faad->srcpad, outbuf);
+          if (ret != GST_FLOW_OK)
+            goto out;
+        } else
+          gst_buffer_unref (outbuf);
       }
     }
   }
@@ -1291,9 +1418,6 @@ next:
 
 out:
 
-  if (caps)
-    gst_caps_unref (caps);
-
   gst_buffer_unref (buffer);
   gst_object_unref (faad);
 
@@ -1307,7 +1431,6 @@ init_failed:
     ret = GST_FLOW_ERROR;
     goto out;
   }
-
 init2_failed:
   {
     GST_ELEMENT_ERROR (faad, STREAM, DECODE, (NULL),
@@ -1315,7 +1438,6 @@ init2_failed:
     ret = GST_FLOW_ERROR;
     goto out;
   }
-
 decode_error:
   {
     GST_ELEMENT_ERROR (faad, STREAM, DECODE, (NULL),
@@ -1339,7 +1461,7 @@ gst_faad_open_decoder (GstFaad * faad)
 
   conf = faacDecGetCurrentConfiguration (faad->handle);
   conf->defObjectType = LC;
-  /* conf->dontUpSampleImplicitSBR = 1; */
+  conf->dontUpSampleImplicitSBR = 1;
   conf->outputFormat = FAAD_FMT_16BIT;
 
   if (faacDecSetConfiguration (faad->handle, conf) == 0) {
@@ -1353,8 +1475,10 @@ gst_faad_open_decoder (GstFaad * faad)
 static void
 gst_faad_close_decoder (GstFaad * faad)
 {
-  faacDecClose (faad->handle);
-  faad->handle = NULL;
+  if (faad->handle) {
+    faacDecClose (faad->handle);
+    faad->handle = NULL;
+  }
 }
 
 static GstStateChangeReturn
@@ -1368,6 +1492,9 @@ gst_faad_change_state (GstElement * element, GstStateChange transition)
       if (!gst_faad_open_decoder (faad))
         return GST_STATE_CHANGE_FAILURE;
       break;
+    case GST_STATE_CHANGE_READY_TO_PAUSED:
+      gst_segment_init (faad->segment, GST_FORMAT_UNDEFINED);
+      break;
     default:
       break;
   }
@@ -1387,6 +1514,7 @@ gst_faad_change_state (GstElement * element, GstStateChange transition)
       faad->prev_ts = GST_CLOCK_TIME_NONE;
       faad->bytes_in = 0;
       faad->sum_dur_out = 0;
+      faad->error_count = 0;
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
       gst_faad_close_decoder (faad);
@@ -1405,7 +1533,7 @@ gst_faad_change_state (GstElement * element, GstStateChange transition)
 static gboolean
 plugin_init (GstPlugin * plugin)
 {
-  return gst_element_register (plugin, "faad", GST_RANK_PRIMARY, GST_TYPE_FAAD);
+  return gst_element_register (plugin, "faad", GST_RANK_SECONDARY, GST_TYPE_FAAD);
 }
 
 GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
